/*
 * F_Pattern_KnockCome.h
 *
 *  Created on: 21. nov. 2021
 *      Author: oyvindteig
 */


#ifndef F_PATTERN_KNOCKCOME_H_
    #define F_PATTERN_KNOCKCOME_H_

    /*
    State machine for the KnockCome (Knock-Come, Knock_Come) pattern

    The knock-come pattern is described in the documentation of the code.
    See usage in the files and in the data-flow diagram below

    It is also described in these blog notes:
       [The "knock-come" deadlock free pattern]
           httsp://oyvteig.blogspot.no/2009/03/009-knock-come-deadlock-free-pattern.html
       [My Beep-BRRR notes - Decoupling Slave_task and Master_task - Implementation D]
           https://www.teigfam.net/oyvind/home/technology/219-my-beep-brrr-notes/#implementation_d

    But here's an example as well. Purpose: Slave_task wants to send data to Master_task:

    CHAN_SIGNAL_SM_A carries the "KNOCK" data-less "message"
    > Slave_task initiates every communication with first sending a signal (no data)
    > on the signal channel (no data, it never blocks) to Master_task. This is the "knock". ("AB" means from Slave_task to Master_task)

    CHAN_SIGNAL_MS_A carries "COME" message
    > When the Master_task is ready to take full data it sends a "come" message back. The sender receives this "come".
    ("MS" means from Master_task to Slave_task)

    CHAN_SYNCH_SM_A carries the data
    > Master_task expects to see only the data in return, so it can wait for it in the "next line" after Come.
    > Slave_task then immediately, by contract, in the "next line" must send off the data set it originally wanted to.
    > Slave_task, after it sent knock, must be able to handle any message from Master_task, because they may be other data, since
    > Master_task is allowed to send on CHAN_SIGNAL_MS_A *any* time

    Two processes can initiale sending to each other this way,and they will never deadlock.
    But it's enough to use knock-come only one way. Therefore Master_task
    only needs CHAN_SIGNAL_MS_A to send anything, at any time (and some times it's the knock message).
    */

    typedef enum {           // NEEDS
        KC_TYP_MS_NONE_DATA, // Master sends spontaneous data to Slave, not part of knock-come scheme
        KC_TYP_SM_KNOCK,     // Slave to Master (not necessary since anything on this streaning chan makes sense)
        KC_TYP_MS_COME,      // Master sends "come!" to Slave, no piggy.backed data
        KC_TYP_MS_COME_DATA, // Master sends "come!" to Slave, but also includes spontaneous piggy-backed data
        KC_TYP_SM_DATA       // Slave sends data to Master. Knock-Come Sequence finished
    } KnockCome_Message_Type_e;

    typedef struct {
        KnockCome_Message_Type_e KnockCome_Message_Type; // KC_TYP_SM_KNOCK only
    } ch_ba_knock_t;

    typedef enum {
        // We don't need a KNOCKCOME_KNOCKSEND_PENDING_TO_SEND_KNOCK_A since we can always send immediately,
        // since it's a SIGNAL-type asynch non-blocking sending
        //
                                                   //               NEXT STATE
        KC_STATE_SLAVE_SENT_DATA_NOW_READY = 0x1A, // Also INIT --> KC_STATE_SLAVE_SENT_KNOCK
        KC_STATE_SLAVE_SENT_KNOCK          = 0x1B, //           --> KC_STATE_SLAVE_GOT_COME
                                                   //               After this we are dependent on how long the master may be busy: KC_THROUGHPUT_TAG
        KC_STATE_SLAVE_GOT_COME            = 0x1C, //           --> KC_STATE_SLAVE_SENT_DATA_NOW_READY
                                                   //
        KC_STATE_MASTER_GOT_DATA_NOW_READY = 0x2A, // Also INIT --> KC_STATE_MASTER_GOT_KNOCK
        KC_STATE_MASTER_GOT_KNOCK          = 0x2B, //           --> KC_STATE_MASTER_SENT_COME
        KC_STATE_MASTER_SENT_COME          = 0x2C, //           --> KC_STATE_MASTER_GOT_DATA_NOW_READY
        //
    } KnockCome_State_e;

    extern KnockCome_State_e
    Slave_Set_KnockCome_State // The callee TASK starts with KNOCK and later SENDS data
    (
        const KnockCome_State_e PresentState,
        const KnockCome_State_e NewState
    );


    extern KnockCome_State_e
    Master_Set_KnockCome_State // The callee TASK responds with COME and then RECEIVES
    (
        const KnockCome_State_e PresentState,
        const KnockCome_State_e NewState
    );

    #define DEBUG_F_PATTERN_KNOCKCOME 0 // 0 or 1

    // Usage:
    // SLAVE_SET_KNOCKCOME_STATE  (PresentState,NewState)
    // MASTER_SET_KNOCKCOME_STATE (PresentState,NewState)
    //
    #if (DEBUG_F_PATTERN_KNOCKCOME == 1)
        #define SLAVE_SET_KNOCKCOME_STATE(PresentState,NewState)  PresentState = Slave_Set_KnockCome_State (PresentState,NewState)
        #define MASTER_SET_KNOCKCOME_STATE(PresentState,NewState) PresentState = Master_Set_KnockCome_State(PresentState,NewState)
    #else
        // DEBUG_PRINT_F_PATTERN_KNOCKCOME not tested here, so no print in this case!
        //
        #define SLAVE_SET_KNOCKCOME_STATE(PresentState,NewState)  PresentState = NewState
        #define MASTER_SET_KNOCKCOME_STATE(PresentState,NewState) PresentState = NewState
    #endif
#else
    #error NESTED .h FILE. BUT IT NEEDS:
    /*
    #include "_globals.h"
    */
#endif /* F_PATTERN_KNOCKCOME_H_ */
